# OALABS
#
# Much of this code was copied from or heavily influnced by Mario Vilas 
# winappdebug project. I am including the full winappdebug copyright notice
# and a link to the project on GitHub.
#
# https://github.com/MarioVilas/winappdbg
#
#
#
# Copyright (c) 2009-2020, Mario Vilas
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import ctypes
# Imports from pywin32
import win32api
import win32security
# Local imports
from . import win32types
from . import win32utils



#===============================================================================
# BOOL WINAPI StackWalk64(
#  __in      DWORD MachineType,
#  __in      HANDLE hProcess,
#  __in      HANDLE hThread,
#  __inout   LPSTACKFRAME64 StackFrame,
#  __inout   PVOID ContextRecord,
#  __in_opt  PREAD_PROCESS_MEMORY_ROUTINE64 ReadMemoryRoutine,
#  __in_opt  PFUNCTION_TABLE_ACCESS_ROUTINE64 FunctionTableAccessRoutine,
#  __in_opt  PGET_MODULE_BASE_ROUTINE64 GetModuleBaseRoutine,
#  __in_opt  PTRANSLATE_ADDRESS_ROUTINE64 TranslateAddress
# );
#===============================================================================
def StackWalk64(MachineType, 
                hProcess, 
                hThread, 
                StackFrame, 
                ContextRecord,
                ReadMemoryRoutine = None,
                FunctionTableAccessRoutine = None, 
                GetModuleBaseRoutine = None,
                TranslateAddress = None):

    _StackWalk64 = ctypes.windll.dbghelp.StackWalk64
    _StackWalk64.argtypes = [win32types.DWORD, 
                            win32types.HANDLE, 
                            win32types.HANDLE, 
                            win32types.LPSTACKFRAME64, 
                            win32types.PVOID,
                            win32types.PREAD_PROCESS_MEMORY_ROUTINE64,
                            win32types.PFUNCTION_TABLE_ACCESS_ROUTINE64,
                            win32types.PGET_MODULE_BASE_ROUTINE64,
                            win32types.PTRANSLATE_ADDRESS_ROUTINE64]
    _StackWalk64.restype = bool

    pReadMemoryRoutine = None
    if ReadMemoryRoutine:
        pReadMemoryRoutine = win32types.PREAD_PROCESS_MEMORY_ROUTINE64(ReadMemoryRoutine)
    else:
        pReadMemoryRoutine = ctypes.cast(None, win32types.PREAD_PROCESS_MEMORY_ROUTINE64)

    pFunctionTableAccessRoutine = None
    if FunctionTableAccessRoutine:
        pFunctionTableAccessRoutine = win32types.PFUNCTION_TABLE_ACCESS_ROUTINE64(FunctionTableAccessRoutine)
    else:
        pFunctionTableAccessRoutine = ctypes.cast(None, win32types.PFUNCTION_TABLE_ACCESS_ROUTINE64)

    pGetModuleBaseRoutine = None
    if GetModuleBaseRoutine:
        pGetModuleBaseRoutine = win32types.PGET_MODULE_BASE_ROUTINE64(GetModuleBaseRoutine)
    else:
        pGetModuleBaseRoutine = ctypes.cast(None, win32types.PGET_MODULE_BASE_ROUTINE64)

    pTranslateAddress = None
    if TranslateAddress:
        pTranslateAddress =  win32types.PTRANSLATE_ADDRESS_ROUTINE64(TranslateAddress)
    else:
        pTranslateAddress = ctypes.cast(None, win32types.PTRANSLATE_ADDRESS_ROUTINE64)

    if ContextRecord is None:
        raise ValueError("ContextRecord cannot be None")
    try:
        pContextRecord = win32types.PCONTEXT(ContextRecord)
    except:
        pContextRecord = win32types.PWOW64_CONTEXT(ContextRecord)

    #this function *DOESN'T* set last error [GetLastError()] properly most of the time.
    ret = _StackWalk64(MachineType, hProcess, hThread, ctypes.byref(StackFrame),
                       pContextRecord, pReadMemoryRoutine,
                       pFunctionTableAccessRoutine, pGetModuleBaseRoutine,
                       pTranslateAddress)

    return ret